/* SCCSID = %W% %E% */
/**************************************************************************
 *
 * SOURCE FILE NAME =  S506EXT.H
 *
 * DESCRIPTIVE NAME =  IBM1S506.ADD - Adapter Driver for ST506/IDE DASD
 *
 *
 * Copyright : COPYRIGHT IBM CORPORATION, 1991, 1992
 *             LICENSED MATERIAL - PROGRAM PROPERTY OF IBM
 *             REFER TO COPYRIGHT INSTRUCTION FORM#G120-2083
 *             RESTRICTED MATERIALS OF IBM
 *             IBM CONFIDENTIAL
 *
 * VERSION = V2.0
 *
 * DATE
 *
 * DESCRIPTION : Data External References
 *
 * Purpose:  This file provides external references for global
 *           data contained in S506DATA.C
 *
 *
 *
 *
 * CHANGE ACTIVITY =
 *   DATE      FLAG        APAR   CHANGE DESCRIPTION
 *   --------  ----------  -----  --------------------------------------
 *   06/21/94  @V87325     87325  Added Suspend/Resume Logic
 *   09/16/95  @V98451            Added Resource Manager
 *   12/16/94  @V108555           Added support for Ontrack Disk Manager and
 *                                Promise Caching IDE Controllers.
 *   01/15/95  @V109483           ATAPI Slave only checking causes DPT reset
 *                                in DoATAPIPresenseCheck to be ignored.
 *   04/03/95  @V117508   117508  Add #defines to include MCA version of driver
 *   04/21/95  @V119250   119250  Do not load on a CMD 640 system
 *   07/07/95  @V129765  PJ19409  More PCI configuration support, RZ1000.
 *   01/06/96  @V147576           Merged CMD, added HWRESOURCE to fix
 *                                Suspend/Resume for multiple clients.
 *   03/30/96  @V149971           Merged warm docking and warm device
 *                                swapping.
 *   04/10/96  @V151345 SRD       fix support for EXECUTE_ATA passthru requests
 *   05/06/96  @V153732           Initialization traps after code merged
 *   05/17/96  @V153916           Deleted Force Load command line option
 *   05/17/96  @V154573           ISA machines failed to load during initialization
 *   05/24/96  @V155162           Thinkpad docking/swapping IDE update.
 *   05/27/96  @V153620           Kernel suspend/resume processing.
 *   07/29/96  @V159935           Trap-D due to configuring 2nd Triton
 *                                channel when no devices attached.
 *   07/30/96  @V159438           Move PhystoVirt from interrupt to init
 *                                time to fix trap-8, interrupt stack ovfl.
 *   11/21/96  @V172802           Deleted DockII command line support as the
 *                                DockII's IDE bus cannot reliably dock and
 *                                undock.  Therefore we only support DockII
 *                                resident IDE devices acutally detected
 *                                during cold boot.
 *   05/06/97  @V179942           PIIX4/Ultra DMA support.
 *   07/16/97  @V181721           Add CMD Verbose external.
 *   01/14/98  @V192088           Add LastCylinderAdjust variable.
 *   01/21/98  @V192545           /W switch to pause verbose output.
 *   02/10/98  @V193287           Dual PIIX handling.
 *   02/20/98  @V192176           >8.4GB support.
 ****************************************************************************/
/*
 * Edit History for defect # 148406:
 *
 * Edit    Date    Comments
 * ----- --------  -------------------------------------------------------------
 * [001] 05-10-95  Integrated changes from Shishir Shah's WDCIDEOS.ADD V1.01
 *                 driver into latest OS/2 Warp driver. In this file this involved
 *                 adding several references to additional messages. /jlh
 *
 * [002] 06-01-95  Add Bus Master DMA capabilities to merged driver. /jlh
 *      1) Add enhanced informational messages for DMA enabled and type.
 *      2) Add PCI Bus Master IDE DMA support.
 *
 * [003] 08-31-95  Integrated latest IBM driver changes. IBM changes flagged with
 *                 unique IBM change codes as in table above. /jlh
 *
 * [004] 10-11-95  Moved ScratchSGList and added two other structures
 *                 to area of data that remains after initialization
 *                 has completed. /jlh
 * [005] 11-14-95  Merged in latest IBM code changes. /jlh
 *
 * [006] 11-16-95  Add vars to aid in serializing IOCTL requests. /jlh
 *
 * [007] 11-29-95  Add PIIX3_SIDETIM and PIIX3Msgtxt. Remove BM_DMA_BaseAddress./mol
 *
 * [008] 12-21-95 Moved PIIX3Present to global permanent storage here. /jlh
 *
 * [009] 03-06-96  Add variables required for Micro House EzDrive support. /jlh
*/

#include "cmdphdr.h"

/*-------------------------------------------------------------------*/
/*                                                                   */
/*      Static Data                                                  */
/*                                                                   */
/*                                                                   */
/*-------------------------------------------------------------------*/


 extern ACBPTRS      ACBPtrs[MAX_ADAPTERS];                         /*@V147576, modified */
 extern PFN          Device_Help;
 extern HDRIVER      hDriver;                                       /*@V98451*/
 extern USHORT       ADDHandle;
 extern USHORT       cAdapters;
 extern USHORT       cUnits;
 extern USHORT       InitActive;
 extern USHORT       InitIOComplete;
 extern USHORT       BIOSActive;
 extern USHORT       LightState;
 extern USHORT       LightCount;
 extern USHORT       MachineID;
 extern USHORT       Calibrate;
 extern USHORT       XBIOSPresent;                                  /*@V108555*/
 extern USHORT       EzDrivePresent;                                /* [009] */
 extern ULONG        WaitDRQCount;
 extern ULONG        CheckReadyCount;
 extern ULONG        IODelayCount;
 extern ULONG        ElapsedTimerHandle;
 extern ULONG        ReqBlockID;                                    /*@V149971*//* moved @V155162*/
 extern VOID         (FAR *pCardServices)(VOID);                    /*@V129765, moved */
 extern VOID         (FAR *pOEMHLPEntry)(VOID);                     /*@V129765, moved */
 extern USHORT       OemhlpDS;                                      /*@V129765, moved */
 extern HWRESOURCE   HWResource[MAX_ADAPTERS];                      /*@V147576*/

 extern UCHAR        AdapterName[17];  /* Adapter Name ASCIIZ string          */
 extern CHAR         PacRimID[];                                    /*@V155162*/
 extern USHORT       DriverBusType;                                 /*@V154573*/

/*---------------------------------------------------------------*/ /*@V129765*/
/*                                                               */ /*@V129765*/
/*      PCI Configuration Operations                             */ /*@V129765*/
/*                                                               */ /*@V129765*/
/*                                                               */ /*@V129765*/
/*---------------------------------------------------------------*/ /*@V129765*/
                                                                    /*@V129765*/
 extern PCI_CONFIG_OPERATION   PCIConfigOp[MAX_PCI_CONFIG_OPS];     /*@V129765*/
 extern UCHAR                  PCIBaseAddressReg[MAX_PCI_BASE_REGS];/*@V129765*/
 extern PCI_DEVICE             PCIDevice[MAX_PCI_DEVICE_DESCRIPT];  /*@V129765*/
 extern UCHAR                  MaxPCIBuses;                         /*@V129765*/
 extern UCHAR                  PCIMajorVer;                         /*@V129765*/
 extern UCHAR                  PCIMinorVer;                         /*@V129765*/
 extern USHORT ReqCount;                                            /* [006] */
 extern ULONG  Queue;                                               /* [006] */
 extern USHORT PIIX_ModeSets[2][4];                                 /*@V179942*/

 extern UCHAR TimerPool[TIMER_POOL_SIZE];
 extern CHAR           OEMDriverName[];                             /*@V153732*/
 extern CHAR           PCMCIA_DDName[];                             /*@V155162*/
 extern IDCTABLE       DDTable;                                     /*@V153732*/
 extern IDCTABLE       PCMIDC;                                      /*@V155162*/
#ifdef DEBUG                                                        /*@V159438*/
 extern PUSHORT        pIRQNestLevel;                               /*@V159438*/
 extern DEBUG_RECORD   debRec[MAX_DEBUG_RECORDS];                   /*@V159438*/
#endif                                                              /*@V159438*/
 extern ACB   ACBPool[];


/*-------------------------------------------------------------------*/
/*                                                                   */
/*      Initialization Data                                          */
/*                                                                   */
/*                                                                   */
/*-------------------------------------------------------------------*/

 extern NPACB          npACBPool;
 extern CHAR           IDEPnPID[];                                  /*@V153620*/
 extern UCHAR          PnPwork[1024];                               /*@V153620*/
 extern ATBL           AdapterTable[MAX_ADAPTERS];
 extern ULONG          CalibrateTimerHandle;
 extern USHORT         CallWorkerSync;
 extern USHORT         CallWorker;
 extern USHORT         Verbose;
 extern USHORT         ATAPISlaveChk;                               /*@V109483*/
 extern IORB_EXECUTEIO InitIORB;
 extern SCATGATENTRY   ScratchSGList;                               /*@V108555*/
 extern PDDD_PARM_LIST pDDD_Parm_List;
 // Scratch buffers used to get the BIOS Parameter Block            /*@V179486*/
 // during initializaion.                                           /*@V179486*/
 extern BOOT_RECORD           BootRecord;                           /*@V179486*/
 extern BOOT_RECORD           ExtendedBootRecord;                   /*@V179486*/
 extern PARTITION_BOOT_RECORD PartitionBootRecord;                  /*@V179486*/
 extern IDCTABLE       DDTable;
 extern ULONG          RegBlock;                                    /*@V155162*/
 extern USHORT         ClientHandle;                                /*@V155162*/
 // Kernel suspend driver request, stop queueing work flag.         /*@V153620*/
 extern BOOL           OS2Suspend;                                  /*@V153620*/
 extern USHORT         SystemFlags;     /* [002.2] PCI Bus related information flags */

 extern USHORT         LastCylinderAdjust;                          /*@V192088*/

 extern MSGTABLE       InitMsg;
 extern NPSZ           AdptMsgs[];
 extern NPSZ           UnitMsgs[];
 extern NPSZ           MsgSMSOn;
 extern NPSZ           MsgLBAOn;
 extern NPSZ           MsgDMAOn;        /* [001] DMA Enabled message */
 extern NPSZ           MsgDMAtxt;                                   /*@V179942*/
 extern NPSZ           MsgBMOn;                                     /*@V179942*/
 extern NPSZ           MsgTypeBDMAOn;   /* [002.1] DMA Enabled message */
 extern NPSZ           MsgTypeFDMAOn;   /* [002.1] DMA Enabled message */
 extern NPSZ           MsgTypeMDMAOn;   /* [002.2] Bus Master DMA Enabled message */
 extern NPSZ           MsgSGOn;         /* [001] DMA scatter/gather enabled msg */
 extern NPSZ           MsgONTrackOn;                                /*@V108555*/
 extern NPSZ           MsgEzDrive;      /* [016] EzDrive present signon text */
 extern NPSZ           MsgEzDriveFBP;   /* [016] EzDrive present and FBP enabled */
 extern NPSZ           ATAPIMsg;                                     /*@V87325*/
 extern NPSZ           MsgUnitForce;                                /*@V149971*/
 extern NPSZ           MsgACBNotVia;                                /*@V159935*/
 extern NPSZ           MsgNull;
 extern UCHAR          CMD640Msgtxt[];                              /*@V129765*/
 extern UCHAR          RZ1000Msgtxt[];                              /*@V129765*/
 extern UCHAR          IN371FBMsgtxt[];                             /* [00x] */
 extern UCHAR          PIIX3Msgtxt[];                               /* [007] */
 extern UCHAR          PIIX4Msgtxt[];                               /*@V179942*/
 extern UCHAR          ParmErrMsg[];
 extern UCHAR          VersionMsg[];
 extern UCHAR          EBIOSIdString[];                             /*@V192176*/
 extern UCHAR          PromiseIdString[];                           /*@V108555*/
 extern USHORT         PromiseIdLength;                             /*@V108555*/
 extern UCHAR          EzDriveIdString[];                           /* [009] */
 extern USHORT         EzDriveIdLength;                             /* [009] */
 extern UCHAR          ScratchBuf[SCRATCH_BUF_SIZE];
#ifdef MCA                                                          /*@V117508*/
 extern USHORT         SystemType;                                  /*@V117508*/
 extern BYTE           InitABRB1[];                                 /*@V117508*/
#endif                                                              /*@V117508*/

/* Begin [001] */

 extern NPSZ           WDCModel;
 extern UCHAR          ErrMsg1[];
 extern UCHAR          ErrMsg2[];
 extern UCHAR          ErrMsg3[];
 extern UCHAR          ErrMsg4[];

/* End [001] */

/*-------------------------------------------------------------------*/
/*                                                                   */
/*      Drive Type Table                                             */
/*                                                                   */
/*                                                                   */
/*-------------------------------------------------------------------*/

 extern DRIVETYPE      DriveTypeTable[MAX_DRIVE_TYPES];


/*-------------------------------------------------------------------*/
/*                                                                   */
/*      Command Line Parser Data                                     */
/*                                                                   */
/*                                                                   */
/*-------------------------------------------------------------------*/

 extern PSZ          pcmdline1;
 extern PSZ          pcmdline_slash;
 extern PSZ          pcmdline_start;
 extern INT          tokv_index;
 extern INT          state_index;
 extern INT          length;
 extern CHARBYTE     tokvbuf[];
 extern POPT         pend_option;
 extern POPT         ptable_option;
 extern BYTE         *poutbuf1;
 extern BYTE         *poutbuf_end;
 extern CC           cc;
 extern USHORT       outbuf_len;
 extern PBYTE        poutbuf;
 extern OPTIONTABLE  opttable;
 extern USHORT       DrvrNameSize;                                  /*@V98451*/
 extern DRIVERSTRUCT DriverStruct;                                  /*@V98451*/
 extern ADAPTERSTRUCT AdapterStruct;                                /*@V98451*/
 extern DEVICESTRUCT DevStruct;                                     /*@V98451*/
 extern UCHAR        ATAPIDescriptName[];                           /*@V98451*/
 extern UCHAR        DevDescriptNameTxt[];                          /*@V98451*/

/*----------------------------------------------------------------*//*@V192176*/
/*                                                                *//*@V192176*/
/*      Mini-VDM Data                                             *//*@V192176*/
/*                                                                *//*@V192176*/
/*                                                                *//*@V192176*/
/*----------------------------------------------------------------*//*@V192176*/
 extern BOOL         VDMInt13Created;                               /*@V192176*/
 extern BOOL         VDMInt13Active;                                /*@V192176*/
 extern VDMINT13CB   VDMInt13;                                      /*@V192176*/
 extern ULONG        VDMDskBlkID;                                   /*@V192176*/
 extern ULONG        VDMInt13BlkID;                                 /*@V192176*/

/*-------------------------------------------------------------------*/
/*                                                                   */
/*      Verbose Data                                                 */
/*                                                                   */
/*                                                                   */
/*-------------------------------------------------------------------*/

 extern UCHAR        VPCIInfo[];                                    /*@V106915*/
 extern UCHAR        VControllerInfo[];                             /*@V106915*/
 extern UCHAR        VUnitInfo1[];                                  /*@V106915*/
 extern UCHAR        VUnitInfo2[];                                  /*@V106915*/
 extern UCHAR        VModelInfo[];                                  /*@V106915*/
 extern UCHAR        VTimingInfo[];                                 /* [002.2] */
 extern UCHAR        VModelUnknown[];                               /*@V106915*/
 extern UCHAR        VBPBInfo[];                                    /*@V179486*/
 extern UCHAR        VCMDInfo[];                                    /*@V181721*/
 extern UCHAR        VI13Info[];                                    /*@V192176*/
 extern UCHAR        VIDEInfo[];                                    /*@V192176*/
 extern UCHAR        VGeomInfo0[];                                  /*@V192176*/
 extern UCHAR        VGeomInfo1[];                                  /*@V106915*/
 extern UCHAR        VGeomInfo2[];                                  /*@V106915*/
 extern UCHAR        VGeomInfo3[];                                  /*@V106915*/
 extern UCHAR        VBlankLine[];                                  /*@V106915*/
 extern UCHAR        VString[];                                     /* [002.1] */
 extern UCHAR        VULTRADMAString[];                             /*@V179942*/
 extern UCHAR        VDMAString[];                                  /* [002.2] */
 extern UCHAR        VPIOString[];                                  /* [002.2] */
 extern UCHAR        VPauseVerbose0[];                              /*@V192545*/
 extern UCHAR        VPauseVerbose1[];                              /*@V192545*/

/* Verbose information */
#define NO_DEVICES_FOUND 6                                          /*@V179942*/

 extern PIIXES       PIIXes[];                                      /*@V193287*/
