/*static char *SCCSID = "%w% /%e%";*/
/**************************************************************************
 *
 * SOURCE FILE NAME = S506PIIX.C
 *
 * DESCRIPTIVE NAME = IBM1S506.ADD - Adapter Driver for ST506/IDE DASD
 *
 *
 * Copyright : COPYRIGHT IBM CORPORATION, 1991, 1992
 *             LICENSED MATERIAL - PROGRAM PROPERTY OF IBM
 *             REFER TO COPYRIGHT INSTRUCTION FORM#G120-2083
 *             RESTRICTED MATERIALS OF IBM
 *             IBM CONFIDENTIAL
 *
 * VERSION = V2.0
 *
 * DATE
 *
 * DESCRIPTION : Adapter Driver PIIX routines.
 *
 * Purpose:
 *
 *
 *
 *
 *
 * CHANGE ACTIVITY =
 *   DATE      FLAG        APAR   CHANGE DESCRIPTION
 *   --------  ----------  -----  --------------------------------------
 *  06/12/97   @V179942           New file - routines moved from S506INIT.c
 *                                to support PIIX4 and Ultra DMA.
 *  08/28/97   @V182487           Removed global piix level and added second
 *                                piix4 chip for docking.
 *  02/10/98   @V193287           Dual PIIX handling.
 ****************************************************************************/

 #define INCL_NOPMAPI
 #define INCL_DOSINFOSEG                                            /*@V153620*/
 #define INCL_NO_SCB
 #define INCL_INITRP_ONLY
 #define INCL_DOSERRORS
 #include "os2.h"
 #include "dos.h"
 #include "bseerr.h"
 #include "dskinit.h"
 #include <scb.h>
 #include <abios.h>

 #include "devhdr.h"
 #include "iorb.h"
 #include "reqpkt.h"
 #include "dhcalls.h"
 #include "addcalls.h"
 #include "devclass.h"

 #include "s506cons.h"
 #include "s506type.h"
 #include "s506regs.h"
 #include "s506ext.h"
 #include "s506pro.h"
 #include "s506misc.h"
 #include "s506ps2.h"

/*--------------------------------------------*/
/*                                            */
/* SetupPIIX()                                */
/*                                            */
/*                                            */
/*--------------------------------------------*/
VOID SetupPIIX( NPACB npACB, NPATBL npAT )                          /*@V182487*/
{
   USHORT i, j;
   USHORT piix_Level = 0;
   ULONG  PhysAddr;

   for (i=0,npAT = AdapterTable; i < 2 ; i++,npAT++ )
   {
     npACB = npAT->npACB;                /* ptr to next Adapter Control Block */

     /*                                                                @V159935
     ** If there were no units attached, then the ACB has been         @V159935
     ** deallocated, so do no configuration on this IDE channel.       @V159935
     */                                                              /*@V159935*/

     if( (!npACB) || (npACB->cUnits == 0) )
        continue;                                                    /*@V159935*/

     if ( npAT->Flags & ATBF_BM_DMA )    /* if Bus Master DMA on this adapter */
     {
       /*-------------------------------------------*/           /*@V182487*/
       /* Get the proper PIIX level for the channel */           /*VVVVVVVV*/
       /*-------------------------------------------*/
       switch( npAT->PCIInfo.Ident.Device )
       {
          case PIIX_PCIIDE_DEV_ID:
               piix_Level = PIIX;
               break;
          case PIIX3_PCIIDE_DEV_ID:
               piix_Level = PIIX3;
               break;
          case PIIX4_PCIIDE_DEV_ID:
               piix_Level = PIIX4;
               break;
            default:
               piix_Level = NOPIIX;
               break;                                            /*AAAAAAAA*/
       }                                                         /*@V182487*/

       npACB->PIIX_IDEtim  = 0;              /* clear timing value initially */
       npACB->PIIX_SIDEtim = 0;        /* clear Slave timing value initially */

       CalculateAdapterTiming( npACB, piix_Level, npAT->BasePort == FX_PRIMARY );

       /***********************************************/
       /* Save PCI Bus Master DMA controller register */
       /* addresses for use later                     */
       /***********************************************/

       npACB->BMICOM = npAT->BM_BaseAddress;
       npACB->BMISTA = npAT->BM_BaseAddress + 2;
       npACB->BMIDTP = npAT->BM_BaseAddress + 4;

       /*************************************/
       /*      Program the PIIX Chip        */
       /*************************************/

       if( ProgramPIIXChip( npACB, npAT, i, piix_Level )) {
          continue;
       }

       /******************************************/
       /* Configure each unit to proper transfer */
       /* mode for PIO and/or DMA                */
       /******************************************/

       for ( j=0; j < npAT->cUnits; j++ )                            /* [003] *//*@V163508*/
       {
         SelectUnit( npACB, j );             /* select unit */

         if ( (npAT->Unit[j].Status == UTS_OK) &&                              /* [003] */
             !( CheckReady( npACB ) && (npACB->TimerFlags & ACBT_BUSY) ) )   /* check ready status *//* [003][018.2] */
         {
            DISABLE                             /* [003] disable interrupts */

            if ( npAT->BM_Overides & (ATBM_NOBM_UNIT0 << j) )
            {
              npACB->UnitCB[j].Flags &= ~UCBF_BM_DMA;
            }

            #ifdef DMA_DEFAULT_OFF                                              /*@V162458*/
             if ( !(npAT->BM_Overides & (ATBM_BM_UNIT0 << j)) )        /*@V162458*/
             {                                                         /*@V162458*/
               npACB->UnitCB[j].Flags &= ~UCBF_BM_DMA;                 /*@V162458*/
             }                                                         /*@V162458*/
            #endif                                                               /*@V162458*/

            if( npACB->UnitCB[j].Flags & UCBF_ATAPIDEVICE )
            {
              if ( !(npAT->BM_Overides & (ATBM_BM_UNIT0 << j)) ) {
                 npACB->UnitCB[j].Flags &= ~UCBF_BM_DMA;
              }
            }

            /*****************************************/
            /* Set Device DMA mode (ULTRA or MW DMA) */
            /*****************************************/

            if( npACB->UnitCB[j].Flags & UCBF_BM_DMA )
            {
               SetDeviceDMAMode( j, npAT, npACB, piix_Level );
            }

            /*********************************/
            /*      Set Device PIO mode      */
            /*********************************/

            SetDevicePIOMode( j, npAT, npACB );

            npACB->Flags &= ~ACBF_INTERRUPT;
            DevHelp_EOI( npACB->IntLevel );                  /* [003] clear IRQ */
            ENABLE                                   /* [003] enable interrupts */
         }
       }

       SelectUnit( npACB, 0 );               /* reselect unit 0 */
     }
     else
     {
        /**************************/
        /* No Adapter BM DMA Flag */
        /* so disable unit BM DMA */
        /**************************/

        DisableAdapterBMDMA( npAT, npACB );
     }

     /*************************/
     /* Set DMA Capable Bits  */
     /*************************/

     SetDMAcapableBits( npACB );
   }
}

/*--------------------------------------------*/
/* CalculateAdapterTiming                     */
/* ----------------------                     */
/*                                            */
/* Arguments:                                 */
/*      npACB = nptr to Adapter Control Block */
/*                                            */
/* Actions:                                   */
/*                                            */
/*      Picks optimal settings for PIO/DMA    */
/*      transfer for the selected ACB.        */
/*                                            */
/* Returns:                                   */
/*      Nothing                               */
/*                                            */
/*                                            */
/*--------------------------------------------*/

VOID CalculateAdapterTiming( NPACB npACB, USHORT piix_Level, USHORT uCable ) /*[009.2]*/
{
  USHORT      BestDMAMode,BestPIOMode;
  USHORT      DMASetting,PIOSetting;
  USHORT      i,shift_value;
  USHORT      PCIClockIndex = 0;

  /* Matrix for optimum PIO timings */
  /*                                */
  /* The idea of this table is to   */
  /* represent normalized timing    */
  /* modes. Rather than figuring on */
  /* separate tables for PIO/DMA    */
  /* we will normalize all timings  */
  /* to PIO modes (even future DMA  */
  /* modes without corresponding    */
  /* PIO modes can be normalized to */
  /* artificial PIO modes. This can */
  /* then be used to determine the  */
  /* basic timing information for   */
  /* programming into the chip.     */
  /*                                */
  /* Dx = Drive x, Mx = Mode x      */
  /*                                */
  /*      D1M0 D1M1 D1M2 D1M3 D1M4  */
  /* D0M0 0000 0000 0002 0003 0004  */
  /* D0M1 0000 0000 0002 0003 0004  */
  /* D0M2 0200 0200 0202 0003 0004  */
  /* D0M3 0300 0300 0300 0303 0303  */
  /* D0M4 0400 0400 0400 0303 0404  */

  static USHORT BestPIOTable[5][5] =
  {
    { 0x0000, 0x0000, 0x0002, 0x0003, 0x0004 },
    { 0x0000, 0x0000, 0x0002, 0x0003, 0x0004 },
    { 0x0200, 0x0200, 0x0202, 0x0003, 0x0004 },
    { 0x0300, 0x0300, 0x0300, 0x0303, 0x0303 },
    { 0x0400, 0x0400, 0x0400, 0x0303, 0x0404 }
  };

  /**********************************/
  /* Pick correct clock index value */
  /**********************************/

  /* always use 33Mhz timings */

//  if ( SystemFlags & SYSFLG_CLK25 )
//  {
//     PCIClockIndex = 1;                    /* 25 Mhz timings */
//  }

  /****************************************************************/
  /* If >= PIIX3 then determine best PIO/DMA timing for each unit */
  /* For the MASTER unit save the timing in PIIX_IDEtim,          */
  /* and for the SLAVE, save timing in piix3_SIDEtim.             */
  /****************************************************************/

  if( piix_Level >= PIIX3 )                        /* PIIX3 and beyond */
  {
     for( i = 0; i < npACB->cUnits; i++ )          /* For each attached unit */
     {
        /**************************************/
        /* Find optimum interface timing mode */
        /**************************************/

        DeviceTimingMode( i, npACB );

        /************************/
        /* Set PIOsetting value */
        /************************/

        PIOSetting = npACB->UnitCB[i].InterfaceTiming;

        /**********************************/
        /* shift value for piix3_SIDEtim: */
        /* primary - 8 , secondary - 4    */
        /**********************************/
        shift_value = uCable ? 8 : 4;

        /******************************************/
        /* Choose PIIX timing register base value */
        /******************************************/

        TimeRegBaseValue( i, shift_value, PIOSetting, PCIClockIndex, npACB );

        /*************************************************/
        /* Select timing for each Device unit on channel */
        /*************************************************/

        SelectDeviceTiming( i, npACB, piix_Level );

        /***********************************************/
        /* Force compatible mode on PIO if we can run  */
        /* fast DMA but not fast PIO                   */
        /***********************************************/

        ForceComptMode( i, npACB );
     }
  } else                                              /* PIIX less than PIIX3 */
  if( npACB->cUnits == 2 )
  {
     /*************************************/
     /* Choose best timing mode valid     */
     /* for device 0/1 whether PIO or DMA */
     /*************************************/

     DeviceTimingMode( 0, npACB );
     DeviceTimingMode( 1, npACB );

     /***********************************/
     /* Find Optimum timing for Channel */
     /***********************************/

     BestPIOMode = BestPIOTable[npACB->UnitCB[0].InterfaceTiming]
                               [npACB->UnitCB[1].InterfaceTiming];

     npACB->UnitCB[1].InterfaceTiming = BestPIOMode & 0x00FF;
     npACB->UnitCB[0].InterfaceTiming = BestPIOMode >> 8;

     /*******************************************/
     /* Pick highest speed PIO mode for channel */
     /*******************************************/

     if ( npACB->UnitCB[0].InterfaceTiming > npACB->UnitCB[1].InterfaceTiming )
     {
        PIOSetting = npACB->UnitCB[0].InterfaceTiming;
     } else {
        PIOSetting = npACB->UnitCB[1].InterfaceTiming;
     }

     /******************************************/
     /* Choose PIIX timing register base value */
     /******************************************/

     TimeRegBaseValue( npACB->cUnits, 0, PIOSetting, PCIClockIndex, npACB );

     /******************************************/
     /* Select timing for each unit on channel */
     /******************************************/

     SelectDeviceTiming( 0, npACB, piix_Level );
     SelectDeviceTiming( 1, npACB, piix_Level );

     /***********************************************/
     /* Force compatible mode on PIO if we can run  */
     /* fast DMA but not fast PIO                   */
     /***********************************************/

     ForceComptMode( 0, npACB );
     ForceComptMode( 1, npACB );
  }
  else                                /* only 1 attached device (Master only) */
  {
     /**************************************/
     /* Find optimum interface timing mode */
     /**************************************/

     DeviceTimingMode( 0, npACB );

     /******************************************/
     /* Find highest PIO speed for the channel */
     /******************************************/

     PIOSetting = npACB->UnitCB[0].InterfaceTiming;

     /******************************************/
     /* Choose PIIX timing register base value */
     /******************************************/

     TimeRegBaseValue( npACB->cUnits, 0, PIOSetting, PCIClockIndex, npACB );

     /******************************************/
     /* Select timing for ONLY unit on channel */
     /******************************************/

     SelectDeviceTiming( 0, npACB, piix_Level );

     /***********************************************/
     /* Force compatible mode on PIO if we can run  */
     /* fast DMA but not fast PIO                   */
     /***********************************************/

     ForceComptMode( 0, npACB );
  }
}

/*-------------------------------*/
/*                               */
/* DeviceTimingMode()            */
/*                               */
/*-------------------------------*/
VOID DeviceTimingMode( USHORT i, NPACB npACB )
{
   int j;

  /*************************************/
  /* Matrix to match DMA/PIO timings   */
  /*                                   */
  /* Rows are DMA modes,               */
  /* Cols are PIO modes                */
  /* result is PIO mode to program     */
  /*                                   */
  /*           P0 P1 P2 P3 P4          */
  /*        D0 00 00 00 03 04          */
  /*        D1 03 03 03 03 03          */
  /*        D2 04 04 04 03 04          */
  /*************************************/

  static UCHAR MatchPIODMATable[3][5] =
  {
     { 0x00, 0x00, 0x00, 0x03, 0x04 },
     { 0x03, 0x03, 0x03, 0x03, 0x03 },
     { 0x04, 0x04, 0x04, 0x03, 0x04 }
  };

   npACB->UnitCB[i].InterfaceTiming = npACB->UnitCB[i].CurPIOMode;

   if ( npACB->UnitCB[i].Flags & UCBF_BM_DMA )
   {
      j = MatchPIODMATable[npACB->UnitCB[i].CurDMAMode]
                          [npACB->UnitCB[i].InterfaceTiming];

      /**************************************/
      /* if j not equal to Current PIO mode */
      /* else interfacetiming = CurPIOMode  */
      /**************************************/

      if ( j != npACB->UnitCB[i].CurPIOMode )
      {
         if ( j < 3 )          /* if drive can't run at least MODE 3 timings */
         {                                   /* then turn off Bus Master DMA */
            npACB->UnitCB[i].Flags &= ~(UCBF_DMASGMODE | UCBF_BM_DMA);
         }
         else if ( ((j-1) == npACB->UnitCB[i].InterfaceTiming) &&
                   (npACB->UnitCB[i].InterfaceTiming >= 3) )
         {                           /* if only off 1 tick then adjust down */
            npACB->UnitCB[i].InterfaceTiming -= 1;
         }
         else
         {
            npACB->UnitCB[i].CurPIOMode = 0;  /* force PIO to mode 0 timings */
            npACB->UnitCB[i].InterfaceTiming = j;    /* use DMA mode timings */
         }
      }
   }
}

/*-------------------------------*/
/*                               */
/* TimeRegBaseValue()            */
/*                               */
/*-------------------------------*/
VOID TimeRegBaseValue( USHORT i, USHORT j, USHORT PIOSetting,
                                 USHORT PCIClockIndex, NPACB npACB )
{
   USHORT units, device;            /* Note =>i has a dual purpose */

   if (!j)                          /* no shift value implies less than PIIX3 */
   {
      units = i;                    /* Note=>i represents number of deivces */

      if( (PIOSetting == 3) || (PIOSetting == 4) )
      {
        npACB->PIIX_IDEtim = PIIX_ModeSets[PCIClockIndex][(PIOSetting-1)];

        if ( npACB->UnitCB[0].Flags & UCBF_BM_DMA )             /* Master - 0 */
        {
          npACB->UnitCB[0].CurDMAMode = (PIOSetting-2);
        }
        if( units == 2 )                                        /* Slave  - 1 */
        {
           if ( npACB->UnitCB[1].Flags & UCBF_BM_DMA )
           {
             npACB->UnitCB[1].CurDMAMode = (PIOSetting-2);
           }
        }
      } else {                                              /* PIOSetting < 3 */
         npACB->PIIX_IDEtim = PIIX_ModeSets[PCIClockIndex][0];
      }
   } else {
      /******************************/
      /* j implies PIIX3 and beyond */
      /* seperate device timing     */
      /* j represents shift value   */
      /******************************/
      device = i;                        /* Note=> i represents deivce number 0 or 1 */

      if( (PIOSetting == 3) || (PIOSetting == 4) )
      {
        if( device )                                          /* Slave timing */
        {
           npACB->PIIX_SIDEtim |= ( ( PIIX_ModeSets[PCIClockIndex][(PIOSetting-1)]
                           & 0x3000 ) >> ( j+2 ) ) |
           (( PIIX_ModeSets[PCIClockIndex][(PIOSetting-1)] & 0x0300 ) >>  j );

        } else {
           npACB->PIIX_IDEtim = PIIX_ModeSets[PCIClockIndex][(PIOSetting-1)];
        }

        if ( npACB->UnitCB[i].Flags & UCBF_BM_DMA )
        {
          npACB->UnitCB[i].CurDMAMode = (PIOSetting-2);
        }
      } else {                                              /* PIOSetting < 3 */
         if( device )
         {                                                    /* Slave timing */
            npACB->PIIX_SIDEtim |= ( ( PIIX_ModeSets[PCIClockIndex][0] & 0x3000 )
                          >> ( j+2 ) ) |
                      (( PIIX_ModeSets[PCIClockIndex][0] & 0x0300 ) >>  j );
         } else {
            npACB->PIIX_IDEtim = PIIX_ModeSets[PCIClockIndex][0];
         }
      }
   }
}

/*-------------------------------*/
/*                               */
/* SelectDeviceTiming()          */
/*                               */
/*-------------------------------*/
VOID SelectDeviceTiming( USHORT i, NPACB npACB, USHORT piix_Level )
{
   /*****************************************/
   /* if FAST timing on unit then enable it */
   /*****************************************/
   if( npACB->UnitCB[i].InterfaceTiming > 2 )
   {
     if( !i ) {                                                 /* Master - 0 */
        npACB->PIIX_IDEtim |= (ACBX_IDETIM_TIME0 | ACBX_IDETIM_IE0);
     } else {                                                    /* Slave - 1 */
        npACB->PIIX_IDEtim |= (ACBX_IDETIM_TIME1 | ACBX_IDETIM_IE1);

        if (piix_Level >= PIIX3) {             /* Slave Timing Register enable */
           npACB->PIIX_IDEtim |=  ACBX_IDETIM_ITE1;
        }
     }

     /*********************************************/
     /*  Enable Pre-Fetch/Posting if FIXED Disk   */
     /*********************************************/
     if( !(npACB->UnitCB[i].Flags & UCBF_ATAPIDEVICE) )
     {
        if( !i ) {                                       /* Master - 0 */
           npACB->PIIX_IDEtim |= ACBX_IDETIM_PPE0;
        } else {                                         /* Slave  - 1 */
           npACB->PIIX_IDEtim |= ACBX_IDETIM_PPE1;
        }
     }
   }
   else                                       /* slow timings, Disable BM DMA */
   {
     npACB->UnitCB[i].Flags &= ~(UCBF_DMASGMODE | UCBF_BM_DMA);
   }
}

/*-------------------------------*/
/*                               */
/* ForceComptMode()               */
/*                               */
/*-------------------------------*/
VOID ForceComptMode( USHORT i, NPACB npACB )
{
   /********************************************************************/
   /* If device PIO capability is much slower than it's DMA capability */
   /********************************************************************/

   if ( (npACB->UnitCB[i].Flags & UCBF_BM_DMA) &&
        (npACB->UnitCB[i].InterfaceTiming) &&     /* if fast interface timing */
        !(npACB->UnitCB[i].CurPIOMode) )          /* but slow PIO */
   {
      /******************************/
      /* DMA Timing Enable Only bit */
      /******************************/
      if( !i ) {
         npACB->PIIX_IDEtim |= ACBX_IDETIM_DTE0;       /* Master - 0 */
      } else {
         npACB->PIIX_IDEtim |= ACBX_IDETIM_DTE1;       /* Slave  - 1 */
      }
   }
   else                                          /* report actual PIO setting */
   {
      npACB->UnitCB[i].CurPIOMode = npACB->UnitCB[i].InterfaceTiming;
   }
}

/*-------------------------------*/
/*                               */
/* SetDeviceDMAMode()            */
/*                               */
/* For ATA or ATAPI devices      */
/*                               */
/*-------------------------------*/
VOID SetDeviceDMAMode( USHORT j, NPATBL npAT, NPACB npACB, USHORT piix_Level )
{
   /************************/
   /* Multi-Word DMA Setup */
   /************************/

   USHORT Port, Data;
   Port = npAT->BasePort + FI_PWRP;                      /* Features Register */
   Data = FX_SETXFERMODE;

   outp ( Port, Data );                                  /* Set TRANSFER Mode */
   IODelay();

   Port = npAT->BasePort + FI_PSECCNT;
   Data = npACB->UnitCB[j].CurDMAMode;
   Data |= FX_MWDMAMODEX;                               /* MultiWord DMA Mode */

   /*****************************************/
   /*   Check for and set Ultra DMA Mode    */
   /*****************************************/

   if( piix_Level > PIIX3 )
   {
     if( npACB->UnitCB[j].UltraDMAMode )                      /* Ultra Device */
     {
        Data = npACB->UnitCB[j].UltraDMAMode;                /* Get ULTRA DMA */
        Data >>= 1;                                             /* zero based */

        /************************************/
        /* Set CurDMAMode for ULTRA Verbose */
        /************************************/
        npACB->UnitCB[j].CurDMAMode = Data + 3;    /* 5=Mode2 4=Mode1 3=Mode0 */

        Data |= FX_ULTRADMAMODEX;                       /* set ULTRA DMA mode */
     }
   }

   outp ( Port, Data );                                       /* Set DMA Mode */
   IODelay();

   npACB->Flags |= ACBF_INTERRUPT;

   Port = npAT->BasePort + FI_PCMD;                       /* Command Register */
   Data = FX_SETFEAT;

   outp ( Port, Data );                                       /* Set FEATURES */
   IODelay();

   CheckReady( npACB );                      /* do rdychk to clear int & wait */
}

/*-------------------------------*/
/*                               */
/* SetDevicePIOMode()            */
/*                               */
/* For ATA or ATAPI devices      */
/*                               */
/*-------------------------------*/
VOID SetDevicePIOMode( USHORT j, NPATBL npAT, NPACB npACB )
{
   USHORT Port, Data;

   Port = npAT->BasePort + FI_PWRP;                      /* Features Register */
   Data = FX_SETXFERMODE;

   outp ( Port, Data );                                  /* Set TRANSFER mode */
   IODelay();

   Port = npAT->BasePort + FI_PSECCNT;
   Data = npACB->UnitCB[j].CurPIOMode;
   if ( Data )                                           /* If not mode 0 PIO */
   {
     Data |= FX_PIOMODEX;                        /* set flow control PIO mode */
   } else {
     Data = FX_PIOMODE0;                                    /* set PIO mode 0 */
   }

   outp ( Port, Data );                                       /* Set PIO Mode */
   IODelay();

   npACB->Flags |= ACBF_INTERRUPT;

   Port = npAT->BasePort + FI_PCMD;                       /* Command Register */
   Data = FX_SETFEAT;

   outp ( Port, Data );                                       /* Set FEATURES */
   IODelay();

   CheckReady( npACB );               /* wait for ready although we ignore it */
}

/*-------------------------------*/
/*                               */
/* SetDMAcapableBits()           */
/*                               */
/* For ATA or ATAPI devices      */
/*                               */
/*-------------------------------*/
VOID SetDMAcapableBits(NPACB npACB)
{
    USHORT Data, Port;

    Data = 0;                                        /* Zero to no BM devices */

    /***************************************************************/
    /* We assume if ULTRA or MW DMA has been successfully set for  */
    /* the device then we still have the unit BM DMA flag          */
    /***************************************************************/

    if ( npACB->UnitCB[0].Flags & UCBF_BM_DMA )
    {
      Data |= ACBX_BMISTA_D0DMA;                   /* Device 0 DMA capable */
    }
    if ( npACB->UnitCB[1].Flags & UCBF_BM_DMA )
    {
      Data |= ACBX_BMISTA_D1DMA;                   /* Device 1 DMA capable */
    }

    Port = npACB->BMISTA;                          /* address BM Status reg */
    outp( Port, Data );                            /* write bits out */
}

/*-------------------------------*/
/*                               */
/* ProgramPIIXChip()             */
/*                               */
/*-------------------------------*/
USHORT ProgramPIIXChip( NPACB npACB, NPATBL npAT, USHORT i, USHORT piix_Level )
{
   USHORT j, Data;
   ULONG  PCIData, TimingMask;

   /***************************************/
   /* Setup PIIX timing for this channel  */
   /***************************************/

   /*************************************/
   /* Write IDE timing Register         */
   /*************************************/

   if( npACB->PIIX_IDEtim )
   {
      if( WritePCIConfigSpace( &AdapterTable[i].PCIInfo,
                               (UCHAR) ( PCI_CR_BM_IDETIM + ((i == 0)
                                                                   ? 0 : 2) ),
                               (ULONG) npACB->PIIX_IDEtim,
                               2) )
      {
         DisableAdapterBMDMA( npAT, npACB );
         return( 1 );
      }
   }

   /**************************************/
   /* Get Master Latency Timer Value     */
   /**************************************/

   if( ReadPCIConfigSpace( &AdapterTable[i].PCIInfo,
                           PCI_CR_BM_MLT,
                           &PCIData,
                           1                        ) )
   {
      DisableAdapterBMDMA( npAT, npACB );
      return( 1 );
    }

   if( PCIData < 0x10 )
   {
      PCIData = 0x20;

      /*************************************/
      /* Write Master Latency Timer Value  */
      /*************************************/

      if( WritePCIConfigSpace( &AdapterTable[i].PCIInfo,
                               PCI_CR_BM_MLT,
                               PCIData,
                               1                        ) )
      {
         DisableAdapterBMDMA( npAT, npACB );
         return( 1 );
      }
   }

   /*************************************/
   /* Read Bus Master Command Register  */
   /*************************************/

   if( ReadPCIConfigSpace( &AdapterTable[i].PCIInfo,
                           PCIREG_COMMAND,
                           &PCIData,
                           2                        ) )
   {
      DisableAdapterBMDMA( npAT, npACB );
      return( 1 );
   }

   /*************************************/
   /* Enable Bus Master IDE Function    */
   /*************************************/

   if( WritePCIConfigSpace( &AdapterTable[i].PCIInfo,
                            PCIREG_COMMAND,
                            PCIData | PCI_CMD_BME,
                            2                       ) )
   {
      DisableAdapterBMDMA( npAT, npACB );
      return( 1 );
   }

   /******************************************************/
   /* Allocate a page of physical memory and map it to   */
   /* this ACB's Bus Master Scatter/Gather list pointer. */
   /* Disable Bus Master DMA if this allocation fails.   */
   /******************************************************/

   if( AllocateBMSGL( npACB ) )                                  /*@V159438*/
   {                                                             /*@V159438*/
      DisableAdapterBMDMA( npAT, npACB );
      return( 1 );
   }                                                             /*@V159438*/

   /*************************************/
   /* Setup PIIX SLAVE Timing Register. */
   /*************************************/

   /**************************************************/
   /* NOTE: PIIX3 and above support seperate timings */
   /**************************************************/

   if( npACB->PIIX_SIDEtim )
   {
      if( WritePCIConfigSpace( &AdapterTable[i].PCIInfo,
                               PCI_CR_BM_SIDETIM,
                               (ULONG) npACB->PIIX_SIDEtim,
                               1                        ) )
      {
         DisableAdapterBMDMA( npAT, npACB );
         return( 1 );
      }
   }

   /*************************************************/
   /*               Ultra DMA Setup                 */
   /*************************************************/

   /*************************************/
   /* PIIX4 and above support ULTRA DMA */
   /*************************************/

   if( piix_Level > PIIX3 )
   {
      /**********************************/
      /* Get ULTRA DMA Control Register */
      /**********************************/

      if( ReadPCIConfigSpace( &AdapterTable[i].PCIInfo,
                              PCI_CR_BM_SDMACTL,
                              &PCIData,
                              1                        ) )
      {
         DisableAdapterBMDMA( npAT, npACB );
         return( 1 );
      }

      /*******************************************/
      /* Determine what drives support Ultra DMA */
      /*******************************************/

      for( j=0; j < npACB->cUnits; j++ )
      {
         Data = ( j + 1 );
         if( i ) {
            Data <<= 2;                  /* shift for Secondary Controller */
         }
         if( npACB->UnitCB[j].UltraDMAMode )                 /* Ultra Device */
         {
            PCIData |= Data;                                 /* Turn On Bit  */
         } else {
            PCIData &= ~Data;                                /* Turn Off Bit */
         }
      }

      /************************************/
      /* Write ULTRA DMA Control Register */
      /************************************/

      if( WritePCIConfigSpace( &AdapterTable[i].PCIInfo,
                               PCI_CR_BM_SDMACTL,
                               PCIData,
                               1                        ) )
      {
         DisableAdapterBMDMA( npAT, npACB );
         return( 1 );
      }

      /*********************************/
      /* Get ULTRA DMA Timing Register */
      /*********************************/

      if( ReadPCIConfigSpace( &AdapterTable[i].PCIInfo,
                              PCI_CR_BM_SDMATIM,
                              &PCIData,
                              2                        ) )
      {
         DisableAdapterBMDMA( npAT, npACB );
         return( 1 );
      }

      /*************************************/
      /* Determine Ultra DMA drive timings */
      /*************************************/

      for( j=0; j < npACB->cUnits; j++ )
      {
         if( npACB->UnitCB[j].UltraDMAMode )                  /* Ultra Device */
         {
            Data = npACB->UnitCB[j].UltraDMAMode;
            Data >>= 1;                                          /* Zero Based */
            TimingMask = 0x0003;
            if( j ) {
               Data <<= 4;                               /* Shift for Drive 1 */
               TimingMask <<= 4;
            }
            if( i ) {
               Data <<= 4;                  /* Shift for Secondary Controller */
               TimingMask <<= 4;
            }
            PCIData &= ~TimingMask;         /* Clear timing bits for drive */
            PCIData |= Data;                /* Set new timing bits         */
         }
      }

      /***********************************/
      /* Write ULTRA DMA Timing Register */
      /***********************************/

      if( WritePCIConfigSpace( &AdapterTable[i].PCIInfo,
                               PCI_CR_BM_SDMATIM,
                               PCIData,
                               2                        ) )
      {
         DisableAdapterBMDMA( npAT, npACB );
         return( 1 );
      }
   }

   /**********************************/
   /* End of Ultra DMA Settings      */
   /**********************************/

   return( 0 );                                           /* Success */
}

/*-------------------------------*/
/*                               */
/* DisableAdapterBMDMA()         */
/*                               */
/*-------------------------------*/
VOID DisableAdapterBMDMA( NPATBL npAT, NPACB npACB )
{
   npAT->Flags &= ~ATBF_BM_DMA;                   /* Adapter DMA Flag */

   npACB->UnitCB[0].Flags &= ~UCBF_BM_DMA;           /* Unit 0 DMA */

   if( npACB->cUnits == 2 )
   {
      npACB->UnitCB[1].Flags &= ~UCBF_BM_DMA;        /* Unit 1 DMA */
   }
}

/*-------------------------------*/                                /*@V159438 */
/*                               */                                /*@VVVVVVV */
/* AllocateBMSG()                */
/*                               */
/*-------------------------------*/

/***********************************************************************/
/* The Bus Master Scatter/Gather table is allocated in physical        */
/* memory and does not move, it is reused for each BM DMA operation.   */
/*                                                                     */
/* 1. Allocate a GDT selector.                                         */
/* 2. Allocate physical memory for scatter/gather table.               */
/* 3. Map the physical memory to the GDT selector.                     */
/*                                                                     */
/* Disable Bus Master DMA if any part of this allocation fails.        */
/***********************************************************************/
BOOL NEAR AllocateBMSGL( NPACB npACB )
{
   ULONG           PhysAddr;

   /*
   ** Initialize the outputs that live in the ACB.
   */
   npACB->pBMDMA_SGL        = NULL;
   npACB->BMDMA_SGListSize  = 0;
   npACB->BMDMA_SGListCount = 0;


   if( DevHelp_AllocGDTSelector( &npACB->BMDMA_GDT, 1 ) )
   {
      /* Error - disable DMA on this adapter */
      return( TRUE );
   }

   if( DevHelp_AllocPhys( MR_4K_LIMIT, 0, &PhysAddr ) )
   {
      /* Error - disable DMA on this adapter */
      DevHelp_FreeGDTSelector( npACB->BMDMA_GDT );
      return( TRUE );
   }

   /*
   ** Calculate useable size of scatter/gather table. Restrictions are:
   ** 1) Cannot span a 4KB boundary
   ** 2) must be ULONG aligned
   **
   ** store resulting physical address, size in bytes of useable area,
   ** and number of descriptors permitted.
   */

   if( (npACB->BMDMA_SGListSize = (PhysAddr & (MR_4K_LIMIT-1))) <=
                                  (MR_4K_LIMIT/2) )
   {
      /* Larger half is before 4KB boundary */

      npACB->BMDMA_SGList = (PhysAddr & ~3L); /* ULONG align it */
      if ( PhysAddr & 3 )
      {
         npACB->BMDMA_SGList += 4;         /* keep in boundaries */
      }
      npACB->BMDMA_SGListSize = MR_4K_LIMIT - npACB->BMDMA_SGListSize;
      npACB->BMDMA_SGListSize &= ~3L;     /* adjust size for ULONGs */
      if ( PhysAddr & 3 )
      {
         npACB->BMDMA_SGListSize -= 4;     /* keep in boundaries */
      }
   }
   else
   {
      /* Begin [012] - larger half is after 4KB boundary */

      npACB->BMDMA_SGList = (PhysAddr - npACB->BMDMA_SGListSize) + MR_4K_LIMIT;
      npACB->BMDMA_SGListSize &= ~3L;     /* adjust size for ULONGs */
      /* End [012] */
   }

   if( DevHelp_PhysToGDTSelector( npACB->BMDMA_SGList,
                                  npACB->BMDMA_SGListSize,
                                  npACB->BMDMA_GDT ) )
   {
      /* Error - disable DMA on this adapter */
      DevHelp_FreePhys( PhysAddr );
      DevHelp_FreeGDTSelector( npACB->BMDMA_GDT );
      return( TRUE );
   }

   /*
   ** It worked OK, so save pointer to Scatter/Gather list in ACB.
   */
   SELECTOROF(npACB->pBMDMA_SGL) = npACB->BMDMA_GDT;
   OFFSETOF(npACB->pBMDMA_SGL) = 0;
   npACB->BMDMA_SGListCount = npACB->BMDMA_SGListSize / 8;

   return( FALSE );                                     /*AAAAAAAA*//*AAAAAAAA*/
}                                                       /*@V159438*//*@V179942*/
                                                                    /*@V193287*/
/*----------------------------------*/
/*                                  */
/*                                  */
/*----------------------------------*/
VOID  DeterminePIIX()
{
   UCHAR    i = 0;
   UCHAR    Row;
   UCHAR    AnyPIIXFound = 0;

   #define  MAXPIIXSEARCH 6

   /*--------------------------------------------------------------*/
   /* PIIXList is an array of PIIX chips that will be searched for */
   /* at Index 0 and Index 1.  If detected we will fill in the     */
   /* Array PIIXes with the PIIX information for that chip.        */
   /* Note: AdapterTable is just used for placeholder for info at  */
   /* first and then is filled in later in CkandValidateChannels   */
   /*--------------------------------------------------------------*/

   static USHORT PIIXList[3][2] =
   {
      { PCID_INTEL82371FB,  PIIX_PCIIDE_DEV_ID },      /* PIIX  */
      { PCID_PIIX3       , PIIX3_PCIIDE_DEV_ID },      /* PIIX3 */
      { PCID_PIIX4       , PIIX4_PCIIDE_DEV_ID }       /* PIIX4 */
   };

   for ( i=0; i< MAXPIIXSEARCH; i++ )
   {
      if( i & 0x04)             /* ROW Calculation */
      {
         Row = 2;
      } else if( i & 0x02) {
         Row = 1;
      } else {
         Row = 0;
      }

      PCIDevice[ PIIXList[Row][0] ].Ident.Index = (i & 0x01);  /* 0 or 1 ONLY */

      if( !( PresenseCheckPCI( &AdapterTable[0],
                               PCIDevice[ PIIXList[Row][0] ] )))
      {
         PIIXes[i].PCIInfo.Present     = TRUE;
         PIIXes[i].PCIInfo.Ident       = PCIDevice[PIIXList[Row][0]].Ident;
         PIIXes[i].npPCIDeviceMsg      = PCIDevice[PIIXList[Row][0]].npDeviceMsg;
         PIIXes[i].PCIInfo.Ident.Device= PIIXList[Row][1];
         PIIXes[i].PCIInfo.FirstConfigRec = AdapterTable[0].PCIInfo.FirstConfigRec;
         PIIXes[i].PCIInfo.BusNum      = AdapterTable[0].PCIInfo.BusNum;
         PIIXes[i].PCIInfo.DevFunc     = AdapterTable[0].PCIInfo.DevFunc;
         PIIXes[i].PCIInfo.DevFunc    &= 0xF8;   /* mask off function number */
         PIIXes[i].PCIInfo.DevFunc    |= 0x01;          /* IDE is function 1 */
         AnyPIIXFound = 1;
      }
   }

   if( AnyPIIXFound )                                     /* Any PIIX Present */
   {
      CkandValidateChannels();
   }
}

/*----------------------------------*/
/*                                  */
/*                                  */
/*----------------------------------*/
VOID CkandValidateChannels()
{
   UCHAR     i, j;
   UCHAR     AtLeast1Primary   = 0;
   UCHAR     AtLeast1Secondary = 0;
   UCHAR     ConfigReg;
   ULONG     PhysAddr;
   NPATBL    npAT;

   for( i=0; i<MAXPIIXSEARCH; i++)
   {
      if( PIIXes[i].PCIInfo.Present == TRUE )
      {
          AdapterTable[0].PCIInfo = PIIXes[i].PCIInfo;

         /*----------------------------------------*/
         /* Ck 4 Primary and/or Secondary Channels */
         /*----------------------------------------*/
         for ( j = 0; j < 2; j++ )
         {
            ConfigReg = (PCI_CR_BM_IDETIM+(2*j));

            if( !(ReadPCIConfigSpace( &AdapterTable[0].PCIInfo,
                                      ConfigReg,
                                      &PhysAddr,
                                      (USHORT) 2           ) ))
            {
               /*--------------------------*/
               /* Check Channel Enable Bit */
               /*--------------------------*/
               if( PhysAddr & IDETIM_IDE )
               {
                  PIIXes[i].Channels |= (1+j);
                  if( !j ) {
                     AtLeast1Primary++ ;
                  } else {
                     AtLeast1Secondary++ ;
                  }
               }
            }
         }
      }
   }

   /*----------------------------------*/
   /* Validate Controllers             */
   /*----------------------------------*/
   TurnOffBM( FX_PRIMARY   );
   TurnOffBM( FX_SECONDARY );

   if( AtLeast1Primary )
   {                                                   /* ValidateControllers */
      if( AtLeast1Primary == 1)
      {
         npAT = AdapterTable;
         CopyChannelPCIinfo( npAT, FX_PRIMARY );
         GetBMBaseAddress( FX_PRIMARY );

         if( AtLeast1Secondary == 1)
         {
            npAT++;
            CopyChannelPCIinfo( npAT, FX_SECONDARY );
            GetBMBaseAddress( FX_SECONDARY );
         }
      }
   }
}

/*----------------------------------*/
/*                                  */
/*                                  */
/*----------------------------------*/
VOID CopyChannelPCIInfo( NPATBL npAT, USHORT channel )
{
   USHORT i, Channel_Bit;

   if( channel == FX_PRIMARY )
   {
      Channel_Bit = 0x01;                        /* PRIMARY   */
   } else {
      Channel_Bit = 0x02;                        /* SECONDARY */
   }

   for( i=0; i<MAXPIIXSEARCH; i++ )
   {
      if( PIIXes[i].PCIInfo.Present == TRUE )
      {
         if( PIIXes[i].Channels & Channel_Bit )
         {
            npAT->PCIInfo        = PIIXes[i].PCIInfo;
            npAT->npPCIDeviceMsg = PIIXes[i].npPCIDeviceMsg;
            break;
         }
      }
   }
}

/*----------------------------------*/
/*                                  */
/*                                  */
/*----------------------------------*/
VOID  GetBMBaseAddress( USHORT channel )
{
   ULONG  PhysAddr;
   USHORT i;

   if( channel == FX_PRIMARY ) {
     i = 0;
   }else {
     i = 1;
   }

   /*--------------------------------------------*/
   /* Bus Master Interface Base Address Register */
   /*--------------------------------------------*/
   if( ReadPCIConfigSpace( &AdapterTable[i].PCIInfo,
                           PCI_CR_BM_IBA,
                           &PhysAddr,
                           4                        ) )
   {
      /*-------------------------------------------*/
      /* Turn off PCI bus mastering on the channel */
      /*-------------------------------------------*/
      TurnOffBM( channel );

   } else {

      /*---------------------------*/
      /* BM Interface Base Address */
      /*---------------------------*/
      AdapterTable[i].BM_BaseAddress=(((USHORT)(PhysAddr & 0x0000FFF0L))+(i*8)) ;
   }
}

/*----------------------------------*/
/*                                  */
/*                                  */
/*----------------------------------*/
USHORT TurnOffBM( USHORT channel )
{
   if( channel == FX_PRIMARY )
   {
      AdapterTable[ 0 ].PCIInfo.Present = FALSE;
   } else {
      AdapterTable[ 1 ].PCIInfo.Present = FALSE;
   }
}

/*----------------------------------*/
/*                                  */
/*                                  */
/*----------------------------------*/
USHORT  AdjustIfPIIXFound( USHORT SystemType )
{
  USHORT    System_Type;
  ULONG     TritonRev = 0;
  USHORT    i;

  static    PCI_INFO  PciInfo;

  System_Type = SystemType;                       /* Get existing system type */

  for (i = 0; i < 2; i++ )                        /* Turn on if PCI present */
  {
     if( AdapterTable[i].PCIInfo.Present == TRUE )
     {
        /*------------------------*/
        /* Default for valid PIIX */
        /*------------------------*/
        System_Type |= SDMA_TYPE_F;             /* PCI so must have type F */
        System_Type |= SDMA_SG;                          /* scatter/gather */
        System_Type |= SDMA_BusMaster;                   /* Bus Master DMA */
     }
  }

  for (i = 0; i < 2; i++ )
  {
     if( AdapterTable[i].PCIInfo.Present == TRUE )
     {
        if( (AdapterTable[i].PCIInfo.Ident.Device == PIIX_PCIIDE_DEV_ID )
         && (AdapterTable[i].PCIInfo.Ident.Revision < 2) )
        {
           /*-----------------------------------*/
           /* Rev is not capable of proper DMA. */
           /*-----------------------------------*/
           System_Type &= ~( SDMA_BusMaster | SDMA_SG );
        }
        if( AdapterTable[i].PCIInfo.Ident.Device == PIIX_PCIIDE_DEV_ID )
        {
           /*-------------------------------------------*/
           /* Check for the ORION chip.                 */
           /*-------------------------------------------*/
           setmem( (PBYTE)&PciInfo, 0, sizeof(PciInfo) );

           if( !FindPCIDevice( &PciInfo, PCIDevice[PCID_INTEL_ORION] ) )
           {
              /*-----------------------------------------*/
              /* Found Orion chipset, check revision id. */
              /*-----------------------------------------*/
              if( ReadPCIConfigSpace( &PciInfo, 8, &TritonRev, 1 ) )
              {
                 System_Type &= ~( SDMA_BusMaster | SDMA_SG );
              }
              else if( TritonRev < 4 )
              {
                 System_Type &= ~( SDMA_BusMaster | SDMA_SG );
              }
           }
        }
     }
  }
  return( System_Type );                                            /*@AAAAAAA*/
}                                                                   /*@V193287*/

