/*static char *SCCSID = "%w% %e%";*/
/**************************************************************************
 *
 * SOURCE FILE NAME = FL2GEO.C
 *
 * DESCRIPTIVE NAME = IBM2FLPY.ADD - Adapter Driver for ABIOS Diskette
 *
 *
 * Copyright : COPYRIGHT IBM CORPORATION, 1991, 1992
 *             LICENSED MATERIAL - PROGRAM PROPERTY OF IBM
 *             REFER TO COPYRIGHT INSTRUCTION FORM#G120-2083
 *             RESTRICTED MATERIALS OF IBM
 *             IBM CONFIDENTIAL
 *
 * VERSION = V2.0
 *
 * DATE
 *
 * DESCRIPTION : Routines to implement the geometry commands.
 *
 *
 *
 * CHANGE ACTIVITY =
 *  DATE      FLAG        APAR   CHANGE DESCRIPTION
 *  --------  ----------  -----  --------------------------------------
 *  05/05/94  @V82589     82589  Initial delta
 *  06/28/94  @V86153     86153  On some model 90's the floppy controller
 *                               needs retries to perform an accurate media
 *                               check.  The disable_retry bit was originally
 *                               set, so as to speed up the driver in
 *                               recognizing unformatted media.
 *  11/08/94  J-d79396           Support 1.2MB Media R/W/F Featuer.
 *
 *
 ****************************************************************************/


#define INCL_NOBASEAPI
#define INCL_NOPMAPI
#include <os2.h>
#include <dhcalls.h>
#include <strat2.h>     /* needed to keep reqpkt.h happy */
#include <reqpkt.h>
#include <scb.h>        /* needed to keey abios.h happy */
#include <abios.h>
#include <iorb.h>
#include <addcalls.h>
#include "fl2def.h"
#include "fl2proto.h"
#include "fl2data.h"


/*****************************************************************************/
/*                                                                           */
/*   Routine     : Determine Media                                           */
/*                                                                           */
/*   Description : This routine starts a media determination sequence:       */
/*                                                                           */
/*                 Check the Changeline                                      */
/*                                                                           */
/*                                                                           */
/*****************************************************************************/

VOID NEAR DetermineMedia( PCFN DetermineComplete  )
{
   NPABRB_DSKT_READCHGLINE pRB = (NPABRB_DSKT_READCHGLINE)RequestBlock;

   Drive[pHeadIORB->UnitHandle].DetermineComplete = DetermineComplete;

   pRB->abrbh.Function  = ABFC_DSKT_READ_CHGSIGNAL;
   pRB->abrbh.Unit      = pHeadIORB->UnitHandle;
   pRB->abrbh.RC        = ABRC_START;
   pRB->Reserved_1      = 0;     /* +16H reserved field  */


   CompletionRoutine = DetermineChkChgComplete;
   Retry = 0;
   Stage = ABIOS_EP_START;
   NextStage();
}

/*****************************************************************************/
/*                                                                           */
/*   Routine     : DetermineChkChgComplete                                   */
/*                                                                           */
/*   Description : If the media has not changed, complete if:                */
/*                 - Previous media state = Unformatted                      */
/*                 - Previous media state = Known                            */
/*                 Else read verify RBA 0 (CHS=0,0,1)                        */
/*                                                                           */
/*****************************************************************************/

VOID FAR DetermineChkChgComplete()
{
   NPABRB_DSKT_READCHGLINE pRB = (NPABRB_DSKT_READCHGLINE)RequestBlock;
   USHORT                        ChgLine = TRUE;
   USHORT                        PrevMotorState;


   /*---------------------------------------------------------*/
   /* Collect the results from the changeline check. If       */
   /* the drive does not support changeline, then use         */
   /* the previous motor state, i.e. if the motor was started */
   /* previously, then the media was not changed.             */
   /*---------------------------------------------------------*/

   PrevMotorState = Drive[pHeadIORB->UnitHandle].Flags.MotorOn;

   Drive[pHeadIORB->UnitHandle].Flags.MotorOn = TRUE;

   if ( pRB->abrbh.RC == ABRC_COMPLETEOK )
      {

         if ( !(pRB->ChangeLineStatus & CHANGELINE_ACTIVE) )
            {
               ChgLine = FALSE;
            }
      }
   else
      {
         if ( PrevMotorState )
            {
               ChgLine = FALSE;
            }
      }

   /*-------------------------------------------------*/
   /* If there was no media change, then report the   */
   /* previous media state.                           */
   /*-------------------------------------------------*/

   if ( !ChgLine )
      {
        if ( Drive[pHeadIORB->UnitHandle].Flags.MediaNotFormatted )
           {
              pHeadIORB->ErrorCode = IOERR_MEDIA_NOT_FORMATTED;
              (*Drive[pHeadIORB->UnitHandle].DetermineComplete)();
              return;
           }
        if ( !Drive[pHeadIORB->UnitHandle].Flags.UnknownMedia )
           {
              if( ShouldSetDevParams)                                           //@IBM J-d79396(A)
              {                                                                 //@IBM J-d79396(A)
                  NPABRB_DSKT_SETDEVPARMS pRB = (NPABRB_DSKT_SETDEVPARMS)RequestBlock;//@IBM J-d79396(A)
                  pRB->abrbh.Function = ABFC_SET_DEVICE_PARMS;                  //@IBM J-d79396(A)
                  pRB->abrbh.Unit     = pHeadIORB->UnitHandle;                  //@IBM J-d79396(A)
                  pRB->Reserved_1     = 0;                                      //@IBM J-d79396(A)
                  pRB->BlockSize      = CurrentBlockSize;                       //@IBM J-d79396(A)
                                                                                //@IBM J-d79396(A)
                  pRB->abrbh.RC       = ABRC_START;                             //@IBM J-d79396(A)
                                                                                //@IBM J-d79396(A)
                  CompletionRoutine   = SetDevParamsForRWV;                     //@IBM J-d79396(A)
                  Retry = 0;                                                    //@IBM J-d79396(A)
                  Stage = ABIOS_EP_START;                                       //@IBM J-d79396(A)
                  NextStage();                                                  //@IBM J-d79396(A)
                  return;                                                       //@IBM J-d79396(A)
              }                                                                 //@IBM J-d79396(A)
              pHeadIORB->Status &= ~IORB_ERROR;
              pHeadIORB->ErrorCode = 0;
              (*Drive[pHeadIORB->UnitHandle].DetermineComplete)();
              return;
           }

      }

   /*-------------------------------------------------*/
   /* Issue a read verify so ABIOS can figure out the */
   /* media geometry.                                 */
   /*-------------------------------------------------*/

   {
//@IBM J-d79396(D)      NPABRB_DSKT_RWV pRB = (NPABRB_DSKT_RWV)RequestBlock;
      NPABRB_DETERMINE_MEDIA pRB = (NPABRB_DETERMINE_MEDIA)RequestBlock;

      Drive[pHeadIORB->UnitHandle].Flags.UnknownMedia      = TRUE;
      Drive[pHeadIORB->UnitHandle].Flags.MediaNotFormatted = FALSE;
      Drive[pHeadIORB->UnitHandle].Flags.LogicalMedia      = FALSE;

      /* Verify the first sector on the diskette */
      pRB->abrbh.Function       = ABFC_DSKT_DETERMINE_MEDIA;                    //@IBM J-d79396(A)
      pRB->abrbh.Unit           = pHeadIORB->UnitHandle;                        //@IBM J-d79396(A)
      pRB->RSpeed               = 1;                            //Try All rpm   //@IBM J-d79396(A)
      pRB->Reserved_1           = 0;                                            //@IBM J-d79396(A)
      pRB->Cylinder             = 0;                                            //@IBM J-d79396(A)
      pRB->Head                 = 0;                                            //@IBM J-d79396(A)
      pRB->abrbh.RC             = ABRC_START;                                   //@IBM J-d79396(A)
      CompletionRoutine         = DetermineMedia2;                              //@IBM J-d79396(A)
      CurrentBlockSize          = 2;   // Set Default Block Size 512Bytes/Sector//@IBM J-d79396(A)
      ShouldSetDevParams        = 0;                                            //@IBM J-d79396(A)
//@IBM J-d79396(D)      pRB->abrbh.Function = ABFC_DSKT_VERIFY;
//@IBM J-d79396(D)      pRB->abrbh.Unit     = pHeadIORB->UnitHandle;
//@IBM J-d79396(D)      pRB->Cylinder       = 0;
//@IBM J-d79396(D)      pRB->Head           = 0;
//@IBM J-d79396(D)      pRB->Sector         = 1;
//@IBM J-d79396(D)      pRB->cSectors       = 1;
//@IBM J-d79396(D)      pRB->abrbh.RC       = ABRC_START;
//@IBM J-d79396(D)      pRB->Reserved_2     = 0;     /* +16H reserved field  */
//@IBM J-d79396(D)      pRB->Reserved_3     = 0;     /* +1EH reserved field  */
//@IBM J-d79396(D)
//@IBM J-d79396(D)      CompletionRoutine = DetermineVerifyComplete;
      Retry = 0;
      Stage = ABIOS_EP_START;
      NextStage();
   }

}

/*Japan Only******************************************************************/ //@IBM J-d79396(A)
/*                                                                           */ //@IBM J-d79396(A)
/*   Routine     : SetDevParamsForRWV                                        */ //@IBM J-d79396(A)
/*                                                                           */ //@IBM J-d79396(A)
/*   Description :                                                           */ //@IBM J-d79396(A)
/*                                                                           */ //@IBM J-d79396(A)
/*                                                                           */ //@IBM J-d79396(A)
/*                                                                           */ //@IBM J-d79396(A)
/*****************************************************************************/ //@IBM J-d79396(A)
                                                                                //@IBM J-d79396(A)
VOID FAR SetDevParamsForRWV()                                                   //@IBM J-d79396(A)
{                                                                               //@IBM J-d79396(A)
      NPABRB_DSKT_SETDEVPARMS pRB = (NPABRB_DSKT_SETDEVPARMS)RequestBlock;      //@IBM J-d79396(A)
                                                                                //@IBM J-d79396(A)
      if( pRB->abrbh.RC == ABRC_COMPLETEOK )                                    //@IBM J-d79396(A)
      {                                                                         //@IBM J-d79396(A)
          pHeadIORB->Status &= ~IORB_ERROR;                                     //@IBM J-d79396(A)
          pHeadIORB->ErrorCode = 0;                                             //@IBM J-d79396(A)
      }                                                                         //@IBM J-d79396(A)
      else                                                                      //@IBM J-d79396(A)
          pHeadIORB->ErrorCode = TranslateErrorCode( pRB->abrbh.RC );           //@IBM J-d79396(A)
      (*Drive[pHeadIORB->UnitHandle].DetermineComplete)();                      //@IBM J-d79396(A)
      return;                                                                   //@IBM J-d79396(A)
}                                                                               //@IBM J-d79396(A)
/*Japan Only******************************************************************/ //@IBM J-d79396(A)
/*                                                                           */ //@IBM J-d79396(A)
/*   Routine     : DetermineMedia2                                           */ //@IBM J-d79396(A)
/*                                                                           */ //@IBM J-d79396(A)
/*   Description :                                                           */ //@IBM J-d79396(A)
/*                                                                           */ //@IBM J-d79396(A)
/*                                                                           */ //@IBM J-d79396(A)
/*                                                                           */ //@IBM J-d79396(A)
/*****************************************************************************/ //@IBM J-d79396(A)
                                                                                //@IBM J-d79396(A)
VOID FAR DetermineMedia2()                                                      //@IBM J-d79396(A)
{                                                                               //@IBM J-d79396(A)
      NPABRB_DETERMINE_MEDIA pRB = (NPABRB_DETERMINE_MEDIA)RequestBlock;        //@IBM J-d79396(A)
                                                                                //@IBM J-d79396(A)
      if ( pRB->abrbh.RC == ABRC_UNSUPPORTED_FUNCTION )                         //@IBM J-d79396(A)
          EndDetermineMedia1();                                                 //@IBM J-d79396(A)
      else if ( pRB->abrbh.RC == ABRC_COMPLETEOK )                              //@IBM J-d79396(A)
      {                                                                         //@IBM J-d79396(A)
          if( pRB->BlockSize != 2)                                              //@IBM J-d79396(A)
          {                                                                     //@IBM J-d79396(A)
              ShouldSetDevParams = 1;                                           //@IBM J-d79396(A)
              CurrentBlockSize   = pRB->BlockSize;                              //@IBM J-d79396(A)
          }                                                                     //@IBM J-d79396(A)
          EndDetermineMedia2();                                                 //@IBM J-d79396(A)
      }                                                                         //@IBM J-d79396(A)
      else                                                                      //@IBM J-d79396(A)
      {                                                                         //@IBM J-d79396(A)
          /* Error case                                         */              //@IBM J-d79396(A)
          /* Just after boot, meida change line is active.      */              //@IBM J-d79396(A)
          /* MEDIA_CHANGED information from ABIOS can be ignore.*/              //@IBM J-d79396(A)
          /* Try again.                                         */              //@IBM J-d79396(A)
          if (pRB->abrbh.RC == ABRC_DSKT_MEDIA_CHANGED)                         //@IBM J-d79396(A)
              Drive[pHeadIORB->UnitHandle].Flags.UnknownMedia = TRUE;           //@IBM J-d79396(A)
          pHeadIORB->ErrorCode = TranslateErrorCode( pRB->abrbh.RC );           //@IBM J-d79396(A)
          (*Drive[pHeadIORB->UnitHandle].DetermineComplete)();                  //@IBM J-d79396(A)
      }                                                                         //@IBM J-d79396(A)
}                                                                               //@IBM J-d79396(A)
/*Japan Only******************************************************************/ //@IBM J-d79396(A)
/*                                                                           */ //@IBM J-d79396(A)
/*   Routine     : EndDetermineMedia1                                        */ //@IBM J-d79396(A)
/*                                                                           */ //@IBM J-d79396(A)
/*   Description :                                                           */ //@IBM J-d79396(A)
/*                                                                           */ //@IBM J-d79396(A)
/*                                                                           */ //@IBM J-d79396(A)
/*                                                                           */ //@IBM J-d79396(A)
/*****************************************************************************/ //@IBM J-d79396(A)
                                                                                //@IBM J-d79396(A)
VOID FAR EndDetermineMedia1()                                                   //@IBM J-d79396(A)
{                                                                               //@IBM J-d79396(A)
      NPABRB_DSKT_RWV pRB = (NPABRB_DSKT_RWV)RequestBlock;                      //@IBM J-d79396(A)
      pRB->abrbh.Function = ABFC_DSKT_VERIFY;                                   //@IBM J-d79396(A)
      pRB->abrbh.Unit     = pHeadIORB->UnitHandle;                              //@IBM J-d79396(A)
      pRB->Cylinder       = 0;                                                  //@IBM J-d79396(A)
      pRB->Head           = 0;                                                  //@IBM J-d79396(A)
      pRB->Sector         = 1;                                                  //@IBM J-d79396(A)
      pRB->cSectors       = 1;                                                  //@IBM J-d79396(A)
      pRB->abrbh.RC       = ABRC_START;                                         //@IBM J-d79396(A)
      pRB->Reserved_2     = 0;     /* +16H reserved field  */                   //@IBM J-d79396(A)
      pRB->Reserved_3     = 0;     /* +1EH reserved field  */                   //@IBM J-d79396(A)
                                                                                //@IBM J-d79396(A)
      ShouldSetDevParams  = 0;                                                  //@IBM J-d79396(A)
      CompletionRoutine   = DetermineVerifyComplete;                            //@IBM J-d79396(A)
      Retry = 0;                                                                //@IBM J-d79396(A)
      Stage = ABIOS_EP_START;                                                   //@IBM J-d79396(A)
      NextStage();                                                              //@IBM J-d79396(A)
}                                                                               //@IBM J-d79396(A)
/*Japan Only******************************************************************/ //@IBM J-d79396(A)
/*                                                                           */ //@IBM J-d79396(A)
/*   Routine     : EndDetermineMedia2                                        */ //@IBM J-d79396(A)
/*                                                                           */ //@IBM J-d79396(A)
/*   Description :                                                           */ //@IBM J-d79396(A)
/*                                                                           */ //@IBM J-d79396(A)
/*                                                                           */ //@IBM J-d79396(A)
/*                                                                           */ //@IBM J-d79396(A)
/*****************************************************************************/ //@IBM J-d79396(A)
                                                                                //@IBM J-d79396(A)
VOID FAR EndDetermineMedia2()                                                   //@IBM J-d79396(A)
{                                                                               //@IBM J-d79396(A)
      NPABRB_DSKT_SETDEVPARMS pRB = (NPABRB_DSKT_SETDEVPARMS)RequestBlock;      //@IBM J-d79396(A)
      pRB->abrbh.Function = ABFC_SET_DEVICE_PARMS;                              //@IBM J-d79396(A)
      pRB->abrbh.Unit     = pHeadIORB->UnitHandle;                              //@IBM J-d79396(A)
      pRB->Reserved_1     = 0;                                                  //@IBM J-d79396(A)
                                                                                //@IBM J-d79396(A)
      pRB->abrbh.RC       = ABRC_START;                                         //@IBM J-d79396(A)
                                                                                //@IBM J-d79396(A)
      CompletionRoutine = DetermineVerifyComplete;                              //@IBM J-d79396(A)
      Retry = 0;                                                                //@IBM J-d79396(A)
      Stage = ABIOS_EP_START;                                                   //@IBM J-d79396(A)
      NextStage();                                                              //@IBM J-d79396(A)
}                                                                               //@IBM J-d79396(A)
/*****************************************************************************/
/*                                                                           */
/*   Routine     : VerifyComplete                                            */
/*                                                                           */
/*   Description : This routine is called when the Verify operation          */
/*                 completes.  If the Verify completed successfully          */
/*                 then call Read Media Parameters.                          */
/*                                                                           */
/*****************************************************************************/

VOID FAR DetermineVerifyComplete()
{
   NPABRB_DSKT_READMEDIAPARMS pRB = (NPABRB_DSKT_READMEDIAPARMS)RequestBlock;

   /*-----------------------------------------------------------*/
   /* If the Read Verify completed O.K., issue ABIOS Read Media */
   /* Parameters to extract the media information from ABIOS    */
   /*-----------------------------------------------------------*/

   if ( pRB->abrbh.RC == ABRC_COMPLETEOK )
      {
         /* Call Read Media Parameters */
         pRB->abrbh.Function  = ABFC_DSKT_READ_MEDIA_PARMS;
         pRB->abrbh.RC        = ABRC_START;
         pRB->Reserved_1      = 0;     /* +16H reserved field  */

         CompletionRoutine = DetermineReadParmsComplete;
         Retry = 0;
         Stage = ABIOS_EP_START;
         NextStage();
      }

   /*----------------------------------------------------------------*/
   /* If the Read Verify failed, and the cause was not due           */
   /* to NOT READY or MEDIA CHANGED, then set the MediaNotFormatted  */
   /* flag                                                           */
   /*----------------------------------------------------------------*/

   else
      {


         Drive[pHeadIORB->UnitHandle].Flags.UnknownMedia = TRUE;

         if ( pRB->abrbh.RC != ABRC_DSKT_MEDIA_CHANGED &&
              pRB->abrbh.RC != ABRC_DSKT_MEDIA_NOT_PRESENT )
            {
               Drive[pHeadIORB->UnitHandle].Flags.MediaNotFormatted = TRUE;
            }

         pHeadIORB->ErrorCode = TranslateErrorCode( pRB->abrbh.RC );
         (*Drive[pHeadIORB->UnitHandle].DetermineComplete)();
      }
}


/*****************************************************************************/
/*                                                                           */
/*   Routine     : ReadMediaParamsComplete                                   */
/*                                                                           */
/*   Description : This routine is called when Read Media Parameters         */
/*                 is complete.  If it completed successfully then           */
/*                 this routine updates the drive media geometry             */
/*                 structure with the new geometry and RWV is called         */
/*                 to do the read, write or verify.                          */
/*                                                                           */
/*****************************************************************************/

VOID FAR DetermineReadParmsComplete()
{
   NPABRB_DSKT_READMEDIAPARMS pRB = (NPABRB_DSKT_READMEDIAPARMS)RequestBlock;
   NPGEOMETRY pGeometry;

   if ( pRB->abrbh.RC == ABRC_COMPLETEOK )
      {
         pGeometry = &(Drive[pHeadIORB->UnitHandle].Geometry[MEDIA]);

         pGeometry->BytesPerSector  = 512;
         if( pRB->SectorsPerTrack == 8)                                         //@IBM J-d79396(A)
             pGeometry->BytesPerSector  = 1024;                                 //@IBM J-d79396(A)
         pGeometry->NumHeads        = pRB->cHeads;
         pGeometry->TotalCylinders  = pRB->cCylinders;
         pGeometry->SectorsPerTrack = pRB->SectorsPerTrack;
         pGeometry->TotalSectors    =
            pRB->cCylinders * pRB->cHeads * pRB->SectorsPerTrack;

         Drive[pHeadIORB->UnitHandle].Flags.UnknownMedia      = FALSE;

         Drive[pHeadIORB->UnitHandle].FormatGap = pRB->FormatGap;
      }

  (*Drive[pHeadIORB->UnitHandle].DetermineComplete)();
}
