/**************************************************************************
 *
 * SOURCE FILE NAME = VFDATA.C
 *
 * DESCRIPTIVE NAME = Virtual Floppy Device Driver Data Declaration
 *
 *
 * Copyright : COPYRIGHT IBM CORPORATION, 1991, 1992
 *             Copyright Microsoft Corporation, 1990
 *             LICENSED MATERIAL - PROGRAM PROPERTY OF IBM
 *             REFER TO COPYRIGHT INSTRUCTION FORM#G120-2083
 *             RESTRICTED MATERIALS OF IBM
 *             IBM CONFIDENTIAL
 *
 * VERSION = V2.0
 *
 * DATE
 *
 * DESCRIPTION   This module contains the VFD's private data.
 *
 *
 * FUNCTIONS
 *
 * ENTRY POINTS:
 *
 * DEPENDENCIES:
 *
 * NOTES
 *
 *
 * STRUCTURES
 *
 * EXTERNAL REFERENCES
 *
 * EXTERNAL FUNCTIONS
 *
 * CHANGE ACTIVITY =
 *  DATE      FLAG        APAR   CHANGE DESCRIPTION
 *  --------  ----------  -----  --------------------------------------
 *
 *  05/09/91              2821  B722821 Int13 reset fails in compaqs
 *  05/29/91              0855  B720855 PowerMeter Hangs
 *  10/24/91              3987  A003987 Multiple Dos Boot doesn't work
 *  12/09/91              0024  B730024 Norton backup, motor is turned off
 *                                      after application turned it on
 *  01/15/92              D1535 1535    Disk Device Driver rewrite
 *  02/06/92              9819  AZ09819 IPE when FSACCESS A:
 *  02/26/92              3026  B733026 F19, need to redetermine media after
 *                                      VDM's been mocking w/. it
 *  03/03/92              3516  B733516 Ctx Hook already armed
 *  08/21/92              24780 24780   Norton Backup changeline bit (3F7h)
 *
 *
 ****************************************************************************/

#include "vfdp.h"
#pragma  BEGIN_GLOBAL_DATA

#ifdef   VDDSTRICT
CHAR szAssertMsg[] = "VFLPY assertion failure in %s, line %d\n";
#endif

#ifdef   VDDDEBUG
CHAR szModule[] = "VFLPY: ";
#endif
HHOOK hhookctxtFlpyIntr;              /* context hook handle for floppy intr */
HHOOK hhookDsktMotorTmr;              /* Dskt motor TimeOut   @V0855         */
HVDM hvdmFloppyOwner = NO_VDM;        /* VDM handle of floppy owner          */
FLAGS flFloppy = 0;                   /* floppy flags                        */
BYTE hvdmFloppyOwner3F2;              /* 3F2 access data   @V2821            */
ULONG motor_timeout = MOTOR_TIMEOUT;  /* Floppy motor timeout   @V0024       */
HVDD hVDDVPIC = NULL;                 /* @VD1535 VDDtoVDD comm with          */
                                      /* VPIC. MASK/UNMASK IRQ6              */
HVDHSEM hsemFloppyOwnership;          /* @VD1535 Block VDMs simultaneously   */
                                      /* accessing floppy                    */
PHVDHSEM phsemFloppyOwnership = & hsemFloppyOwnership;/* pointer to sem      */
                                       /* handle                             */
VDHSEMSTATE semStateOwned;             /* Data area for sem info.            */
REDETERMINEMEDIAPARM RedetermineMediaParmPacket,RedetermineMediaDataPacket;
                                       /* @VD1535 SUSPEND/RESTART IOCTL call */
                                       /* parm packet                        */
HFILE hRedetermineMedia = 0;           /* @V3026                             */
ULONG RedetermineMediaLengthData;      /* @V3026                             */
CHAR AFileName[3] = "A:\0";            /* @V3026                             */

/*
** By default we'll trap all direct hardware accesses. This will allow us
** to handle situations where the program uses the ctrlr in peculiar ways.
** Consider what Norton backup does to check if the changeline bit was active,
** the program selects the corresponding drive and then sits there polling
** 3F7 until the changeline bit is set. This is OK until the 2 sec timeout
** timer comes in and turns the motor off. At this point the controller will
** never return the right status if the dskt is replaced because the drive
** is not selected anymore. If the app selected the drive everytime before
** the check is done then things would be fine, but that'd be too easy and
** we wouldn't have the fun of debugging this type of problems.
** INT 13 accesses can be trusted to be more careful than this and thus we
** don't trap all controller accesses for BIOS functions.
**                       ! norton !
*/

CHAR VfSetIOHookVar = 1;              /* norton                              */
#pragma  END_GLOBAL_DATA
#pragma  BEGIN_SWAP_DATA
HIRQ hirq;                            /* floppy IRQ handle                   */
HHOOK hhookvirrWaitPFD = NULL;        /* VIRR hook handle for waiting PFD    */
HHOOK hhooktimerInUseTimeout;         /* timer hook handle for inuse timeout */
IOH iotFDPort = {                     /* port hook structure */
                 VFPortInput, VFPortOutput, NULL, NULL, NULL };

#ifndef  NOMOTORTRAP
IOH iotFDPortDO = {                   /* port hook for digital output reg. */
                 NULL, VFPortDO, NULL, NULL, NULL };

#endif
IOH iotFDPortNoSupport = {            /* port hook for no PFD support        */
                 VFNoSupportIn, VFNoSupportOut, NULL, NULL, NULL } ;

#pragma  END_SWAP_DATA
#pragma  BEGIN_INSTANCE_DATA

/* start @VD1535*/

HFILE DeviceHandle = 0,               /* Handle used to issue                */
Device_Handle = 0;                    /* DevIOCTL calls.                     */
CHAR FileName[3] = "C:\0",            /* Kernel code to allow IOCTL          */
File_Name[3] = "C:\0";                /* calls w/o valid handle not          */
/* working yet, use C: drive.*/
ULONG ActionTaken;
ULONG LengthParm,LengthData;          /* DevIOCTL call parameters            */
HHOOK hhookctxtMotorResume;           /* for calling IOCTL Resume at         */
/* tasktime.  3516*/
SUSPENDRESTARTPARM SuspendRestartParmPacket;/* SUSPEND/RESTART IOCTL call    */

/* parm packet.*/

/* end @VD1535*/

#pragma  END_INSTANCE_DATA
#pragma  BEGIN_SWAP_INSTANCE
HVDM hvdmThisVDM;                     /* VDM handle                          */
HHOOK hhookvirrWaitInUse;             /* VIRR hook handle for waiting InUse  */

#ifdef   MONITORPORT
BOOL fPortTrapEnabled = TRUE;         /* port trap is enabled                */
#endif
#pragma  END_SWAP_INSTANCE
